/*
  mini escreen boot graphics
  Rene Rebe <rene@exactcode.de>
  
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License as
  published by the Free Software Foundation; either version 2 of the
  License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
  USA. */

/* Minimal image to FB, frame-buffer, blitting. Could also serve as
   a Linux FB How-To low-level pixel filling reference that was tested
   on big-endian, likewise. -ReneR */

#define _XOPEN_SOURCE 500
#include <errno.h>
#include <fcntl.h>
#include <locale.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/ioctl.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <termios.h>
#include <unistd.h>

// #include "eTerm.hh"

#include <linux/fb.h>
#include <linux/vt.h>
#include <sys/mman.h>

#include <iostream>
#include <fstream>

#include <Codecs.hh>
#include <scale.hh>

bool chvt (int n)
{
  int fd = open ("/dev/tty0",O_RDONLY);
  if (fd < 0) fd = open ("/dev/vc/0",O_RDONLY);
  if (fd < 0) fd = open ("/dev/console",O_RDONLY);
  if (fd < 0) return false;
  
  if (ioctl(fd,VT_ACTIVATE,n) ||
      ioctl(fd,VT_WAITACTIVE,n)) {
    close(fd);
    return false;
  }
  
#if 0  
  vt_mode vtmode;
  if (ioctl(fd,VT_GETMODE, &vtmode) == 0) {

    std::cerr << "vtmode: " << (int)vtmode.mode
	      << ", wait: " << (int)vtmode.waitv
	      << ", relsig: " << vtmode.relsig
	      << ", acqsig: " << vtmode.acqsig
	      << std::endl; 
    
  }
#endif
  close(fd);
  return true;
}

bool cursor_off ()
{
  static char* cursoff = "\E[?25l\000";
 
  int fd = open ("/dev/tty0", O_RDWR);
  if (fd < 0) fd = open ("/dev/vc/0", O_RDWR);
	if (fd < 0) fd = open ("/dev/console",O_RDWR);
  if (fd < 0) {
    return false;
  }
  else {
    write (fd, cursoff, strlen (cursoff));
    close (fd);
    return true;
  }
}

int main (int argc, char* argv[])
{
  char* file = "t2.xpm";

  if (argc > 1)
    file = argv[1];

  chvt (8) || std::cerr << "Failed to chvt." << std::endl;
  cursor_off () || std::cerr << "Failed to turn cursor off." << std::endl;
  
  int fb;
  if ((fb = open ("/dev/fb0", O_RDWR)) > 0) {
    fb_fix_screeninfo fb_fix;
    fb_var_screeninfo fb_var;
    
    if (ioctl(fb, FBIOGET_VSCREENINFO, &fb_var) == 0 &&
	ioctl(fb, FBIOGET_FSCREENINFO, &fb_fix) == 0)
      {
	if (false) {
	  std::cerr << "visual: " << fb_fix.visual
		    << "" << std::endl;
	  
	  std::cerr << "width: " << fb_var.xres
		    << ", height: " << fb_var.yres << std::endl;
	}
	
	switch (fb_fix.visual) {
	case FB_VISUAL_TRUECOLOR:
		break;
	case FB_VISUAL_DIRECTCOLOR:
		// setup color maps ...
		{
			int rcols = 1 << fb_var.red.length;
			int gcols = 1 << fb_var.green.length;
			int bcols = 1 << fb_var.blue.length;
			int cols = std::max (std::max (rcols,gcols), bcols);
			uint16_t* rmap = (uint16_t*)malloc (cols * sizeof (rmap[0]));
			uint16_t* gmap = (uint16_t*)malloc (cols * sizeof (gmap[0]));
			uint16_t* bmap = (uint16_t*)malloc (cols * sizeof (bmap[0]));
			if (!rmap || !gmap || !bmap) {
				std::cerr << "Failed to allocate color maps." << std::endl;
				return 1;
			}

			for (int i = 0; i < rcols; ++i)
				rmap[i] = (0xffff/(rcols-1)) * i;
			for (int i = 0; i < gcols; ++i)
				gmap[i] = (0xffff/(gcols-1)) * i;
			for (int i = 0; i < bcols; ++i)
				bmap[i] = (0xffff/(bcols-1)) * i;

			fb_cmap cmap;
			cmap.start = 0;
			cmap.transp = 0;
			cmap.len = cols;
			cmap.red = rmap;	
			cmap.green = gmap;
			cmap.blue = bmap;
			cmap.transp = NULL;

			if (ioctl(fb, FBIOPUTCMAP, &cmap) != 0) {
				std::cerr << "Failed setting new color map." << std::endl;
				return 1;
			}

			free(rmap); free(gmap); free(bmap);
		}
		break;
	default:
		std::cerr << "Unsupported visual: " << fb_fix.visual << std::endl;
		return 1;
	}

        if (false)
        std::cerr << "smem_start: " << fb_fix.smem_start
                  << ", smem_len: " << fb_fix.smem_len
                  << ", xpanstep: " << fb_fix.xpanstep
                  << ", ypanstep: " << fb_fix.xpanstep
                  << ", ywrapstep: " << fb_fix.ywrapstep
                  << ", line_length: " << fb_fix.line_length
                  << ", mmio_start: " << fb_fix.mmio_start
                  << ", memio_len: " <<  fb_fix.mmio_len << std::endl;
	
	int width = fb_var.xres;
	int height = fb_var.yres;
	
	uint8_t* fb_mem = (uint8_t*)
	  mmap (0, fb_fix.smem_len, PROT_READ | PROT_WRITE, MAP_SHARED, fb, 0);
	
	if ((void*)fb_mem == MAP_FAILED) {
	  std::cerr << "Failed to mmap FB." << std::endl;
	  exit (1);
	}
	
	std::ifstream xpm (file);
	Image image;
	if (!ImageCodec::Read (&xpm, image)) {
	  std::cerr << "Failed to read XPM image." << std::endl;
	  exit (1);
	}

	if (false)
	std::cerr << fb_var.red.offset << " " << fb_var.red.length << " "
		  << fb_var.green.offset << " " << fb_var.red.length << " "
		  << fb_var.blue.offset << " " << fb_var.blue.length << std::endl;
	
	scale (image, (double) width  / image.w, (double) height / image.h);
	
	if (true)
	std::cerr << "scale: " << (double) width  / image.w << " "
		  << (double) height / image.h << " - " << image.w << " " << image.h << std::endl;
	
	Image::iterator it = image.begin();
	uint16_t r, g, b;
	r = g = b = 0;
#if 0
	/* just for testing */
	switch (fb_var.bits_per_pixel) {
	case 32:
	case 24:
		fb_var.red.offset = atoi(argv[2]);
		fb_var.green.offset = atoi(argv[3]);
		fb_var.blue.offset = atoi(argv[4]);
		break;
	case 16:
		fb_var.red.offset = atoi(argv[2]);
		fb_var.green.offset = atoi(argv[3]);
		fb_var.blue.offset = atoi(argv[4]);
		break;
	default: ;
	}
#endif

	for (int y = 0; y < height; ++y)
	{
	  uint8_t* dst = fb_mem + y * fb_fix.line_length;
	  for (int x = 0; x < width; ++x)
	  {
	    *it;
	    it.getRGB (&r, &g, &b);
	    
	    unsigned int x =
	      ((uint32_t)r >> (8-fb_var.red.length)) << fb_var.red.offset |
	      ((uint32_t)g >> (8-fb_var.green.length)) << fb_var.green.offset |
	      ((uint32_t)b >> (8-fb_var.blue.length)) << fb_var.blue.offset;

	    switch (fb_var.bits_per_pixel) {
	      case 32:
#if __BYTE_ORDER != __BIG_ENDIAN
		dst[0] = (x>>0) & 0xff;
		dst[1] = (x>>8) & 0xff;
		dst[2] = (x>>16) & 0xff;
		dst[3] = (x>>24) & 0xff;
#else
		dst[0] = (x>>24) & 0xff;
		dst[1] = (x>>16) & 0xff;
		dst[2] = (x>>8)	& 0xff;
		dst[3] = (x>>0)	& 0xff;
#endif
		break;
	      case 24:
#if __BYTE_ORDER != __BIG_ENDIAN
		dst[0] = (x>>0) & 0xff;
		dst[1] = (x>>8) & 0xff;
		dst[2] = (x>>16) & 0xff;

#else
		dst[0] = (x>>16) & 0xff;
		dst[1] = (x>>8)	& 0xff;
		dst[2] = (x>>0)	& 0xff;
#endif
		break;
	      case 16:
	      case 15:
#if __BYTE_ORDER != __BIG_ENDIAN
		dst[0] = (x>>0) & 0xff;
		dst[1] = (x>>8) & 0xff;
#else
		dst[0] = (x>>8)	& 0xff;
		dst[1] = (x>>0)	& 0xff;
#endif
		break;
	      default:
		dst[0] = x	& 0xff;
		break;
	    }
	    
	    dst += (fb_var.bits_per_pixel+7) / 8;
	    ++it;
	  }
	}

	/*
	evas = new EvasCanvas ();
	evas->OutputMethod ("fb");
  
	evas->OutputSize (win_w, win_h);
	evas->OutputViewport (0, 0, win_w, win_h);
	
	Evas_Engine_Info_FB* einfo;
	einfo = (Evas_Engine_Info_FB*) evas->EngineInfo ();
  
	// the following is specific to the engine
	einfo->info.virtual_terminal = 0;
	einfo->info.device_number = 0;
	einfo->info.refresh = 0;
	einfo->info.rotation = 0;
  
	evas->EngineInfo ( (Evas_Engine_Info*)einfo );
  
	// Setup
	evas->FontPathPrepend ("/usr/X11/lib/X11/fonts/TTF/");
	evas->FontPathPrepend ("/usr/X11/lib/X11/fonts/TrueType/");
	evas->FontPathPrepend ("/usr/X11/share/fonts/TTF/");
	evas->FontPathPrepend ("/opt/e17/share/evas/data/");
  
	if (true) {
	  // evas->ImageCache (1024 * 1024);
	  //evas->FontCache (256 * 1024);
	}
	else {
	  //evas->ImageCache (0);
	  //evas->FontCache (0);
	}
	
	image = new EvasImage (*evas);
	image->SmartLoad ("/boot/t2.png");
	image->Layer (0);
	image->Show ();
	image->Resize (win_w, win_h);
	image->ImageFill (0, 0, win_w, win_h);
	
	text = new EvasText (*evas);
	text->Font ("style=Bold", 20);
	text->Color (0xFF, 0xFF, 0xFF, 0xFF);
	text->Text (t);
	text->Layer (5);
	text->Show ();
	text->Move ((win_w - text->Width()) / 2, (win_h - text->Height())/2);
	
	// TODO: shell
	
	evas->Render ();
  
	// pause (); // for now, fall thru - a static screen is enough
	
	delete image;
	delete evas;
	evas = 0;  
	*/
      }
    else
      std::cerr << "Error: Failed to obtain FB fix/var info." << std::endl;
  }
  else
    std::cerr << "Error: Failed to open FB device." << std::endl;
}

#if 0

static int child_pid = 0;
static struct termios ttysave;

#include "TermUtil.hh"
// screen width / height
int sw = 80;
int sh = 25;

  // FB init
  struct termios ntio;
  int ret;
  char buf[8192];
  struct timeval tv;
  int ptyfd, ttyfd;
  char *locale = "", *command = NULL;
  int i;
  char o = ' ';
  int pending = 0;
  
  int win_w = 1;
  int win_h = 1;
  
  eTerm* term;
// evas
  EvasCanvas* evas;
  
  struct fb_var_screeninfo fb_var;
  
  int fb;
  if ((fb = open ("/dev/fb0", O_RDWR)) > 0) {
    if (ioctl(fb, FBIOGET_VSCREENINFO, &fb_var) == 0)
      {
	win_w = fb_var.xres;
	win_h = fb_var.yres;
  
	evas = new EvasCanvas ();
	evas->OutputMethod ("fb");
  
	evas->OutputSize (win_w, win_h);
	evas->OutputViewport (0, 0, win_w, win_h);
  
	Evas_Engine_Info_FB* einfo;
	einfo = (Evas_Engine_Info_FB*) evas->EngineInfo ();
  
	/* the following is specific to the engine */
	einfo->info.virtual_terminal = 0;
	einfo->info.device_number = 0;
	einfo->info.refresh = 0;
	einfo->info.rotation = 0;
  
	evas->EngineInfo ( (Evas_Engine_Info*)einfo );
  
	/* Setup */
	evas->FontPathPrepend ("/usr/X11/lib/X11/fonts/TTF/");
	evas->FontPathPrepend ("/usr/X11/lib/X11/fonts/TrueType/");
	evas->FontPathPrepend ("/usr/X11/share/fonts/TTF/");
	evas->FontPathPrepend ("/opt/e17/share/evas/data/");
  
	if (true) {
	  evas->ImageCache (1024 * 1024);
	  evas->FontCache (256 * 1024);
	}
	else {
	  evas->ImageCache (0);
	  evas->FontCache (0);
	}
  
	term = new eTerm (evas, sw, sh);
	sw = win_w / term->FontWidth();
	sh = win_h / term->FontHeight();
	term->Resize (sw, sh);
  
	for (i = 1 ; i < argc ; ++i)
	  if (argv[i][0] == '-')
	    switch (argv[i][1])
	      {
	      case 'f':
	      case 'l':
		o = argv[i][1];
		break;

	      default:
		printf ("unknown option: %c\n", argv[i][1]);
	      }
	  else
	    switch (o)
	      {
	      case ' ':
		command = argv[i];
		break;

	      case 'l':
		locale = argv[i];
		o = ' ';
		break;
	      }

	setlocale (LC_CTYPE, locale);
  
	tcgetattr (0, &ttysave);

	if (get_ptytty (&ptyfd, &ttyfd)) {
	  perror("can't get a pty");
	  exit(1);
	}
  
	spawn_shell (ptyfd, ttyfd, command == NULL ? "/bin/sh" : command);
  
	evas->Render ();
  
	// signal(SIGHUP, reload_font);
	// TODO?: signal (SIGTERM, sigterm);

	ntio = ttysave;
	ntio.c_lflag &= ~(ECHO|ISIG|ICANON|XCASE);
	ntio.c_iflag = 0;
	ntio.c_oflag &= ~OPOST;
	ntio.c_cc[VMIN] = 1;
	ntio.c_cc[VTIME] = 0;
	ntio.c_cflag |= CS8;
	ntio.c_line = 0;
	tcsetattr (0, TCSAFLUSH, &ntio);

	set_window_size (ttyfd, sw, sh);

	for (;;) {
	  fd_set fds;
	  int max = 0;

	  //if (quit)
	  //  break;
    
	  if (pending)
	    {
	      tv.tv_sec = 0;
	      tv.tv_usec = 0;
	    }
	  else
	    {
	      tv.tv_sec = 10;
	      tv.tv_usec = 100000;
	    }
	  FD_ZERO(&fds);
	  FD_SET(0, &fds);
	  FD_SET(ptyfd, &fds);
	  if (ptyfd > max)
	    max = ptyfd;
	  ret = select (max+1, &fds, NULL, NULL, &tv);
    
	  //if (quit)
	  //  break;
    
	  if (ret == 0 || (ret < 0 && errno == EINTR))
	    {
	      if (pending)
		{
		  pending = 0;
	    
		  term->redraw ();
		  evas->Render ();
		  usleep (10000);
		}
	    }
    
	  if (ret < 0)
	    perror("select");
	  if (FD_ISSET(0, &fds)) {
	    ret = read(0, buf, sizeof(buf));
	    if (ret > 0)
	      write (ptyfd, buf, ret);
	  }
	  else if (FD_ISSET(ptyfd,&fds)) {
	    ret = read(ptyfd, buf, sizeof(buf));
	    if (ret > 0)
	      {
		// output to screen
		term->out (buf, ret);
		pending = 1;
	      }
	  }
	}

	delete evas;
	evas = 0;  
      }
    else
      std::cerr << "Error: Failed to obtain FB size." << std::endl;
  }
  else
    std::cerr << "Error: Failed to open FB device." << std::endl;
  
  /*
    Code-flow drafting:
    FB init
    
    if (FB init error) then
      print FB error
    else
      setup pseudo terminal

      fork
      if child then
        run the terminal emulator
      end
    execve (init, args)

    end

    
   */

  // this is the parent process, but even on
  // FB error we get here and just run the real init
  
  execl ("/bin/bash", "/bin/bash", NULL);
}

#endif
